/*
 * Expectation-Maximisation Algorithm for the Mixture of Gaussians (MOG) model.
 *
 * This is only an API proposition and subjected to further changes.
 * http://en.wikipedia.org/wiki/Expectation%E2%80%93maximization_algorithm
 */

#ifndef _EM_MOG_H
#define _EM_MOG_H

enum {
	EM_MOG_ERROR = -1,
	EM_MOG_SUCCESS,
};
/**
 * constant for calculate normal distribution log(2 * M_PI)
 */
#define EM_MOG_LOG_2_PI 1.8378770664093453

/**
 * constant for calculate normal distribution for 3 dimension data 3 * log(2 * M_PI)
 */
#define M3_EM_MOG_LOG_2_PI 5.513631199228036
/**
 * @struct
 * Mixture of Gaussian context
 */
struct em_mog_context {
	/** number of gaussian in model*/
	int ngauss;
	/** data dimension*/
	int dim;
	/** minimal probability for learning samples*/
	double min_loglike;

	/** threshold of error for end iterations in EM algorithm*/
	double epsilon;
	/** number to be added to the covariance matrix to account for the noise signal*/
	double noise_level;

	/** weigghts, means, covariance matrix*/
	double *weight, *means, *cov_matrix;
};

/**
 * @function
 * @return pointer to struct em_mog_context if OK, else NULL
 */
struct em_mog_context *em_mog_new();

/**
 * @function
 * free memory the memory struct em_mog_context
 * @param ctx : pointer to pointer to struct em_mog_context
 */
void em_mog_free(struct em_mog_context **ctx);

/**
 * @function
 * initialization of the structure em_mog_contex and its fields
 * @param ctx : pointer to struct em_mog_context
 */
void em_mog_init(struct em_mog_context *ctx);

/**
 * @function
 * clear fields of structure em_mog_contetx
 * @param ctx : pointer to struct em_mog_context
 */
void em_mog_clear(struct em_mog_context *ctx);

/**
 * @function
 * Set number of gaussians in mixture.
 * @param ctx    : pointer to struct em_mog_context
 * @param ngauss : number of gaussian in mixture
 */
void em_mog_set_ngaussians(struct em_mog_context *ctx, int ngauss);

/**
 * @function
 * Get number of gaussians in mixture.
 * @param ctx : pointer to struct em_mog_context
 * @return number of gaussian in mixture
 */
int em_mog_get_ngaussians(struct em_mog_context *ctx);

/**
 * @function
 * Set fitted data item dimension.
 * @param ctx : pointer to struct em_mog_context
 * @param dim : data dimention
 */
void em_mog_set_dimension(struct em_mog_context *ctx, int dim);
/**
 * @function
 * Get fitted data item dimension.
 * @param ctx : pointer to struct em_mog_context
 * @return mixture data dimention
 */
int em_mog_get_dimension(struct em_mog_context *ctx);

/**
 * @funtion
 * Set error value for end of iteration process.
 * @param ctx   : pointer to struct em_mog_context
 * @param error : error value for the end of iterations
 */
void em_mog_set_error(struct em_mog_context *ctx, double error);
/**
 * @function
 * Get error value for end of iteration process.
 * @param ctx : pointer to struct em_mog_context
 * @return error value for the end of iterations
 */
double em_mog_get_error(struct em_mog_context *ctx);

/**
 * @funtion
 * Set noice_level.
 * @param ctx   : pointer to struct em_mog_context
 * @param noise_level : noise level
 */
void em_mog_set_noise_level(struct em_mog_context *ctx, double noise_level);
/**
 * @function
 * Get noise_level.
 * @param ctx : pointer to struct em_mog_context
 * @return noise_level
 */
double em_mog_get_noise_level(struct em_mog_context *ctx);

/**
 * @function
 * Set n-th gaussian component weight.
 * @param ctx    : pointer to struct em_mog_context
 * @param nth    : number of gaussian component
 * @param weight : weight of gaussian component
 */
void em_mog_set_nth_weight(struct em_mog_context *ctx, int nth, double weight);
/**
 * @function
 * Get n-th gaussian component weight.
 * @param ctx    : pointer to struct em_mog_context
 * @param nth    : number of gaussian component
 * @return weight of gaussian component
 */
double em_mog_get_nth_weight(struct em_mog_context *ctx, int nth);

/**
 * @function
 * Set n-th gaussian mean.
 * @param ctx : pointer to struct em_mog_context
 * @param nth : number of gaussian component
 * @param mu  : mean of gaussian component
 */
void em_mog_set_nth_mean(struct em_mog_context *ctx, int nth, const double *mu);
/**
 * @function
 * Get n-th gaussian mean.
 * @param ctx : pointer to struct em_mog_context
 * @param nth : number of gaussian component
 * @param mu  : array mean of gaussian component
 */
void em_mog_get_nth_mean(struct em_mog_context *ctx, int nth, double *mu);

/* 
 * Convariance matrix data goes in row-major order (C-style). This is a square
 * matrix of MxM size, where M is the dimension of data item. Dimension is
 * provided by context and not required as an argument.
 * 
 * / a b c \
 * | d e f | = { a, b, c, d, e, f, g, h, i }.
 * \ g h i /
 *
 */

/**
 * @funtion
 * Set n-th gaussian covariance matrix.
 * @param ctx : pointer to struct em_mog_context
 * @param nth : number of gaussian component
 * @param cmx : covariance matrix of gaussian component
 */
void em_mog_set_nth_covariance_matrix(struct em_mog_context *ctx, int nth, const double *cmx);
/**
 * @funtion
 * Get n-th gaussian covariance matrix.
 * @param ctx : pointer to struct em_mog_context
 * @param nth : number of gaussian component
 * @param cmx : array covariance matrix of gaussian component
 */
void em_mog_get_nth_covariance_matrix(struct em_mog_context *ctx, int nth, double *cmx);

/* 
 * Do one step of EM-algorithm, returning logarithm of likelihood function.
 * Data fitted to the MOG model is provided by x and it's size in nx. It's
 * dimension is provided by context and not required as a separate argument.
 */
double em_mog_step(struct em_mog_context *ctx, double *x, int nx);

/**
 * @function
 * Converge the EM-algorithm, repeating EM-steps until logarithm of the
 * likelihood function stops to change. Threshold of the change is given by epsilon.
 * @param ctx     : pointer to struct em_mog_context
 * @param x       : data for the calculation parameters of em_mog_context
 * @param nx      : size of data
 * @param epsilon : error value for the end of iterations
 */
int em_mog_converge(struct em_mog_context *ctx, double *x, int nx, double epsilon);

/**
 * @function
 * Determine if a sample x represents model or not.
 * @param ctx : pointer to struct em_mog_context
 * @param x   : sample for testing
 * @return EM_MOG_SUCCESS - if x represents model and data is valid, EM_MOG_ERROR - else
 */
int em_mog_sample_to_match_model(struct em_mog_context *ctx, double *x);

/**
 * @function
 * Update em_mog_context nth element using new sample data
 * @param ctx  : pointer to struct em_mog_context
 * @param data : sample for updating em_mog_context
 * @param n_model : number of updated components of the model
 */
void em_mog_update_nth_model(struct em_mog_context *ctx, double *data, int n_model);

/**
 * @function
 * Update em_mog_context using new sample data
 * @param ctx  : pointer to struct em_mog_context
 * @param data : sample for updating em_mog_context
 */
void em_mog_update(struct em_mog_context *ctx, double *data);

/**
 * @function
 * Reset em_mog_context
 * sets the identity covariance matrix
 * sets weights to 1 / number of gaussians
 * sets means of the random from data (if data != NULL)
 * @param ctx       : pointer to struct em_mog_context
 * @param data      : sample for reset model
 * @param data_size : data size
 * @note
 * if data == NULL fills the means with zeros
 */
void em_mog_reset(struct em_mog_context *ctx, const double *data, int data_size);


#endif

